# -*- coding: utf-8 -*-


r"""
How to update scripts from version 1.0 to 2.0.

:term:`FiPy` 2.0 introduces several syntax changes from :term:`FiPy` 1.0. We appreciate that
this is very inconvenient for our users, but we hope you'll agree that the new
syntax is easier to read and easier to use. We assure you that this is not
something we do casually; it has been over three years since our last
incompatible change (when :term:`FiPy` 1.0 superceded :term:`FiPy` 0.1).

All examples included with version 2.0 have been updated to use the new syntax,
but any scripts you have written for :term:`FiPy` 1.0 will need to be updated. A
complete listing of the changes needed to take the :term:`FiPy` examples scripts from
version 1.0 to version 2.0 can be found with:

    $ git diff version-1_2 version-2_0 examples/

but we summarize the necessary changes here. If these tips are not sufficient to
make your scripts compatible with :term:`FiPy` 2.0, please don't hesitate to ask for
help on the `mailing list`_.


The following items **must** be changed in your scripts

 * The dimension axis of a :class:`~fipy.variables.variable.Variable` is now first, not last

   >>> x = mesh.getCellCenters()[0]

   instead of

   >>> x = mesh.getCellCenters()[..., 0]

   This seemingly arbitrary change simplifies a great many things in :term:`FiPy`, but
   the one most noticeable to the user is that you can now write

   >>> x, y = mesh.getCellCenters()

   instead of

   >>> x = mesh.getCellCenters()[..., 0]
   >>> y = mesh.getCellCenters()[..., 1]

   Unfortunately, we cannot reliably automate this conversion, but we find that
   searching for "``...,``" and "``:,``" finds almost everything. Please don't
   blindly "search & replace all" as that is almost bound to create more
   problems than it's worth.

   .. note::

      Any vector constants must be reoriented. For instance, in order to offset
      a :class:`~fipy.meshes.mesh.Mesh`, you must write

      >>> mesh = Grid2D(...) + ((deltax,), (deltay,))

      or

      >>> mesh = Grid2D(...) + [[deltax], [deltay]]

      instead of

      >>> mesh = Grid2D(...) + (deltax, deltay)

 * :class:`~fipy.variables.vectorCellVariable.VectorCellVariable` and
   :class:`~fipy.variables.vectorFaceVariable.VectorFaceVariable` no longer exist.
   :class:`~fipy.variables.cellVariable.CellVariable` and
   and :class:`~fipy.variables.faceVariable.FaceVariable` now both inherit from
   :class:`~fipy.varibles.meshVariable.MeshVariable`, which
   can have arbitrary rank. A field of scalars (default) will have ``rank=0``, a
   field of vectors will have ``rank=1``, etc. You should write

   >>> vectorField = CellVariable(mesh=mesh, rank=1)

   instead of

   >>> vectorField = VectorCellVariable(mesh=mesh)

   .. note::

      Because vector fields are properly supported, use vector operations to
      manipulate them, such as

      >>> phase.getFaceGrad().dot((( 0, 1),
      ...                          (-1, 0)))

      instead of the hackish

      >>> phase.getFaceGrad()._take((1, 0), axis=1) * (-1, 1)

 * For internal reasons, :term:`FiPy` now supports
   :class:`~fipy.variables.cellVariable.CellVariable` and
   :class:`~fipy.variables.faceVariable.FaceVariable`
   objects that contain integers, but it is not meaningful to solve a PDE for an
   integer field (:term:`FiPy` should issue a warning if you try). As a result, when given,
   initial values must be specified as floating-point values:

   >>> var = CellVariable(mesh=mesh, value=1.)

   where they used to be quietly accepted as integers

   >>> var = CellVariable(mesh=mesh, value=1)

   If the ``value`` argument is not supplied, the
   :class:`~fipy.variables.cellVariable.CellVariable` will contain floats, as before.

 * The ``faces`` argument to
   :class:`~fipy.boundaryConditions.boundaryCondition.BoundaryCondition` now takes a mask,
   instead of a list of :class:`~fipy.meshes.face.Face` IDs. Now you write

   >>> X, Y = mesh.getFaceCenters()
   >>> FixedValue(faces=mesh.getExteriorFaces() & (X**2 < 1e-6), value=...)

   instead of

   >>> exteriorFaces = mesh.getExteriorFaces()
   >>> X = exteriorFaces.getCenters()[..., 0]
   >>> FixedValue(faces=exteriorFaces.where(X**2 < 1e-6), value=...)

   With the old syntax, a different call to
   :meth:`~fipy.meshes.face.Face.getCenters` had to be made for each set
   of :class:`~fipy.meshes.face.Face` objects. It was also extremely
   difficult to specify boundary conditions that depended both on position in
   space and on the current values of any other :class:`~fipy.variables.variable.Variable`.

   >>> FixedValue(faces=(mesh.getExteriorFaces()
   ...                   & (((X**2 < 1e-6)
   ...                       & (Y > 3.))
   ...                      | (phi.getArithmeticFaceValue()
   ...                         < sin(gamma.getArithmeticFaceValue())))), value=...)

   although it probably could have been done with a rather convoluted (and
   slow!) ``filter`` function passed to ``where``. There no longer are any ``filter``
   methods used in :term:`FiPy`. You now would write

   >>> x, y = mesh.cellCenters
   >>> initialArray[(x < dx) | (x > (Lx - dx)) | (y < dy) | (y > (Ly - dy))] = 1.

   instead of the *much* slower

   >>> def cellFilter(cell):
   ...     return ((cell.center[0] < dx)
   ...             or (cell.center[0] > (Lx - dx))
   ...             or (cell.center[1] < dy)
   ...             or (cell.center[1] > (Ly - dy)))

   >>> positiveCells = mesh.getCells(filter=cellFilter)
   >>> for cell in positiveCells:
   ...     initialArray[cell.ID] = 1.

   Although they still exist, we find very little cause to ever call
   :meth:`~fipy.meshes.mesh.Mesh.getCells`
   or :meth:`fipy.meshes.mesh.Mesh.getFaces`.

 * Some modules, such as :mod:`fipy.solvers`, have been significantly rearranged.
   For example, you need to change

   >>> from fipy.solvers.linearCGSSolver import LinearCGSSolver

   to either

   >>> from fipy import LinearCGSSolver

   or

   >>> from fipy.solvers.petsc.linearCGSSolver import LinearCGSSolver


 * The :func:`numerix.max` and :func:`numerix.min` functions no longer exist. Either
   call :func:`max` and :func:`min` or the :meth:`~fipy.variables.variable.Variable.max`
   and :meth:`~fipy.variables.variable.Variable.min` methods of a
   :class:`~fipy.variables.variable.Variable`.

 * The :mod:`Numeric` module has not been supported for a long time. Be sure to use

   >>> from fipy import numerix

   instead of

   >>> import Numeric

The remaining changes are not *required*, but they make scripts easier to read
and we recommend them. :term:`FiPy` may issue a :exc:`DeprecationWarning` for some cases,
to indicate that we may not maintain the old syntax indefinitely.

 * All of the most commonly used classes and functions in :term:`FiPy` are directly
   accessible in the :mod:`fipy` namespace. For brevity, our examples now start with

   >>> from fipy import *

   instead of the explicit

   >>> from fipy.meshes.grid1D import Grid1D
   >>> from fipy.terms.powerLawConvectionTerm import PowerLawConvectionTerm
   >>> from fipy.variables.cellVariable import CellVariable

   imports that we used to use. Most of the explicit imports should continue to
   work, so you do not need to change them if you don't wish to, but we find our
   own scripts much easier to read without them.

   All of the :mod:`~fipy.tools.numerix` module is now imported into the :mod:`fipy` namespace, so you
   can call :mod:`~fipy.tools.numerix` functions a number of different ways, including:

   >>> from fipy import *
   >>> y = exp(x)

   or

   >>> from fipy import numerix
   >>> y = numerix.exp(x)

   or

   >>> from fipy.tools.numerix import exp
   >>> y = exp(x)

   We generally use the first, but you may see us use the others, and should
   feel free to use whichever form you find most comfortable.

   .. note::

      Internally, :term:`FiPy` uses explicit imports, as is considered
      `best Python practice`_, but we feel that clarity trumps orthodoxy when it
      comes to the examples.

   .. _best Python practice: http://docs.python.org/howto/doanddont.html#from-module-import

 * The function :func:`fipy.viewers.make` has been renamed to
   :func:`fipy.viewers.Viewer`. All of the ``limits`` can now be supplied as direct
   arguments, as well (although this is not required). The result is a more
   natural syntax:

   >>> from fipy import Viewer
   >>> viewer = Viewer(vars=(alpha, beta, gamma), datamin=0, datamax=1)

   instead of

   >>> from fipy import viewers
   >>> viewer = viewers.make(vars=(alpha, beta, gamma),
   ...                       limits={'datamin': 0, 'datamax': 1})

   With the old syntax, there was also a temptation to write

   >>> from fipy.viewers import make
   >>> viewer = make(vars=(alpha, beta, gamma))

   which can be very hard to understand after the fact (``make``? ``make`` what?).

 * A :class:`~fipy.terms.convectionTerm.ConvectionTerm` can now calculate its
   Péclet number automatically, so the ``diffusionTerm`` argument is no longer required

   >>> eq = (TransientTerm()
   ...       == DiffusionTerm(coeff=diffCoeff)
   ...       + PowerLawConvectionTerm(coeff=convCoeff))

   instead of

   >>> diffTerm = DiffusionTerm(coeff=diffCoeff)
   >>> eq = (TransientTerm()
   ...       == diffTerm
   ...       + PowerLawConvectionTerm(coeff=convCoeff, diffusionTerm=diffTerm))

 * An :class:`~fipy.terms.term.implicitSourceTerm.ImplicitSourceTerm` now "knows"
   how to partition itself onto the solution matrix, so you can write

   >>> S0 = mXi * phase * (1 - phase) - phase * S1
   >>> source = S0 + ImplicitSourceTerm(coeff=S1)

   instead of

   >>> S0 = mXi * phase * (1 - phase) - phase * S1 * (S1 < 0)
   >>> source = S0 + ImplicitSourceTerm(coeff=S1 * (S1 < 0))

   It is definitely still advantageous to hand-linearize your source terms, but
   it is no longer necessary to worry about putting the "wrong" sign on the
   diagonal of the matrix.

 * To make clearer the distinction between iterations, timesteps, and sweeps
   (see FAQ :ref:`FAQ-IterationsTimestepsSweeps`)
   the ``steps`` argument to a :class:`~fipy.solvers.solver.Solver` object has
   been renamed ``iterations``.

 * :class:`~fipy.terms.implicitDiffusionTerm.ImplicitDiffusionTerm` has been
   renamed to :class:`~fipy.terms.diffusionTerm.DiffusionTerm`.

.. _mailing list:         http://www.ctcms.nist.gov/fipy/mail.html
"""
__docformat__ = 'restructuredtext'




r"""

To run this example from the base FiPy directory type::

    $ python examples/diffusion/steadyState/mesh1D/tri2Dinput.py

at the command line. A contour plot should appear and the word `finished`
in the terminal.

This example is similar to the example found in
:mod:`examples.diffusion.mesh1D`, however, the `mesh` is a
:class:`fipy.meshes.tri2D.Tri2D` object rather than a
:func:`~fipy.meshes.factoryMeshes.Grid1D` object.

Here, one time step is executed to implicitly find the steady state
solution.  We refine the solver tolerance from the default
:math:`10^{-5}` in order to achieve a good solution.

    >>> eq = DiffusionTerm()
    >>> solver = eq.getDefaultSolver(tolerance=1e-8)
    >>> eq.solve(var, solver=solver)

To test the solution, the analytical result is required. The `x`
coordinates from the mesh are gathered and the length of the domain,
`Lx`, is calculated.  An array, `analyticalArray`, is calculated to
compare with the numerical result,

    >>> x = mesh.cellCenters[0]
    >>> Lx = nx * dx
    >>> analyticalArray = valueLeft + (valueRight - valueLeft) * x / Lx

Finally the analytical and numerical results are compared with a
tolerance of `1e-10`.

    >>> print(var.allclose(analyticalArray))
    True

"""
__docformat__ = 'restructuredtext'

from fipy import input
from fipy import Tri2D, CellVariable, DiffusionTerm, Viewer

nx = 50
dx = 1.

mesh = Tri2D(dx = dx, nx = nx)

valueLeft = 0.
valueRight = 1.
var = CellVariable(name = "solution-variable", mesh = mesh, value = valueLeft)

var.constrain(valueLeft, mesh.facesLeft)
var.constrain(valueRight, mesh.facesRight)

if __name__ == '__main__':
    DiffusionTerm().solve(var)
    viewer = Viewer(vars=var)
    viewer.plot()
    x = mesh.cellCenters[0]
    Lx = nx * dx
    analyticalArray = valueLeft + (valueRight - valueLeft) * x / Lx
    print(var.allclose(analyticalArray))
    input("finished")




__docformat__ = 'restructuredtext'

from fipy.solvers.petsc.petscKrylovSolver import PETScKrylovSolver

__all__ = ["LinearCGSolver", "LinearPCGSolver"]

class LinearCGSolver(PETScKrylovSolver):

    """Interface to the conjugate gradient (:term:`CG`) solver in
    :ref:`PETSC`.
    """

    solver = 'cg'

    def _canSolveAsymmetric(self):
        return False

LinearPCGSolver = LinearCGSolver



"""An object oriented, partial differential equation (PDE) solver

:term:`FiPy` is based on a standard finite volume (FV) approach.  The
framework has been developed in the Materials Science and Engineering
Division (MSED_) and Center for Theoretical and Computational Materials
Science (CTCMS_), in the Material Measurement Laboratory (MML_) at the
National Institute of Standards and Technology (NIST_).

The solution of coupled sets of PDEs is ubiquitous to the numerical
simulation of science problems.  Numerous PDE solvers exist, using a
variety of languages and numerical approaches. Many are proprietary,
expensive and difficult to customize.  As a result, scientists spend
considerable resources repeatedly developing limited tools for
specific problems.  Our approach, combining the FV method and :term:`Python`,
provides a tool that is extensible, powerful and freely available. A
significant advantage to :term:`Python` is the existing suite of tools for
array calculations, sparse matrices and data rendering.

The :term:`FiPy` framework includes terms for transient diffusion,
convection and standard sources, enabling the solution of arbitrary
combinations of coupled elliptic, hyperbolic and parabolic PDEs.  Currently
implemented models include phase field :cite:`BoettingerReview:2002`
:cite:`ChenReview:2002` :cite:`McFaddenReview:2002` treatments of
polycrystalline, dendritic, and electrochemical phase transformations, as
well as drug eluting stents :cite:`Saylor:2011p2794`, reactive wetting
:cite:`PhysRevE.82.051601`, photovoltaics :cite:`Hangarter:2011p2795` and a
level set treatment of the electrodeposition process
:cite:`NIST:damascene:2001`.

.. _MML:                  http://www.nist.gov/mml/
.. _CTCMS:                http://www.ctcms.nist.gov/
.. _MSED:                 http://www.nist.gov/mml/msed/
.. _NIST:                 http://www.nist.gov/
"""
__docformat__ = 'restructuredtext'

import json
import logging
import logging.config
import os
import sys

from fipy.boundaryConditions import *
from fipy.meshes import *
from fipy.solvers import *
from fipy.terms import *
from fipy.tools import *
from fipy.tools import parallelComm
from fipy.tools.logging import environment
from fipy.variables import *
from fipy.viewers import *

from . import _version

# configure logging before doing anything else, otherwise we'll miss things
if 'FIPY_LOG_CONFIG' in os.environ:
    with open(
        os.environ['FIPY_LOG_CONFIG'],
        mode='r',
        encoding="utf-8"  # should be "locale" (PEP 597), needs py>3.9
    ) as config:
        logging.config.dictConfig(json.load(config))

_log = logging.getLogger(__name__)

# __version__ needs to be defined before calling package_info()
__version__ = _version.get_versions()['version']

_fipy_environment = {
    "argv": sys.argv,
    "environ": dict(os.environ),
    "platform": environment.platform_info(),
    "package": environment.package_info()
}

if _log.isEnabledFor(logging.DEBUG):
    try:
        _fipy_environment.update(environment.conda_info())
    except Exception as e:
        _log.error("conda-info: %s", e)

    try:
        _fipy_environment.update(environment.pip_info())
    except Exception as e:
        _log.error("pip-info: %s", e)

    try:
        _fipy_environment.update(environment.nix_info())
        raise ValueError("wow!")
    except Exception as e:
        _log.error("nix-info: %s", e)

_log.debug(json.dumps(_fipy_environment))

# fipy needs to export raw_input whether or not parallel

input_original = input

if parallelComm.Nproc > 1:
    def mpi_input(prompt=""):
        """Replacement for `input` for multiple processes
        """
        parallelComm.Barrier()
        sys.stdout.flush()
        if parallelComm.procID == 0:
            sys.stdout.write(prompt)
            sys.stdout.flush()
            return sys.stdin.readline()

        return ""
    input = mpi_input
else:
    input = input_original

_saved_stdout = sys.stdout


def _serial_doctest_raw_input(prompt):
    """Replacement for `raw_input()` that works in doctests
    """
    _saved_stdout.write("\n")
    _saved_stdout.write(prompt)
    _saved_stdout.flush()
    return sys.stdin.readline()


def doctest_raw_input(prompt):
    """Replacement for `raw_input()` that works in doctests

    This routine attempts to be savvy about running in parallel.
    """
    try:
        parallelComm.Barrier()
        _saved_stdout.flush()
        if parallelComm.procID == 0:
            txt = _serial_doctest_raw_input(prompt)
        else:
            txt = ""
        parallelComm.Barrier()
    except ImportError:
        txt = _serial_doctest_raw_input(prompt)
    return txt


def test(*args, **kwargs):
    r"""
    Test :term:`Fipy`. Equivalent to::

        $ fipy_test --modules

    Use::

        $ fipy_test --help

    for a full list of options. Options can be passed in the same way
    as they are appended at the command line. For example, to test
    :term:`FiPy` with :ref:`Trilinos` and inlining switched on, use

    >>> fipy.test(trilinos=True, inline=True)

    At the command line this would be::

        $ fipy_test --modules --trilinos --inline

    .. note::

        A :command:`fipy_test` option like :option:`--deprecation-errors`
        is equivalent to the :func:`~fipy.test` argument
        ``deprecation_errors``.

    """

    from fipy.tests.test import main

    try:
        main(modules=True, *args, **kwargs)
    except SystemExit:
        pass



__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.variables.cellVariable import CellVariable
from fipy.tools import parallelComm
from .gapFillMesh import GapFillMesh

class TrenchMesh(GapFillMesh):

    """
    The following test case tests for diffusion across the domain.

    >>> cellSize = 0.05e-6
    >>> trenchDepth = 0.5e-6
    >>> boundaryLayerDepth = 50e-6
    >>> domainHeight = 10 * cellSize + trenchDepth + boundaryLayerDepth

    >>> mesh = TrenchMesh(trenchSpacing = 1e-6,
    ...                   cellSize = cellSize,
    ...                   trenchDepth = trenchDepth,
    ...                   boundaryLayerDepth = boundaryLayerDepth,
    ...                   aspectRatio = 1.) # doctest: +GMSH

    >>> import fipy.tools.dump as dump
    >>> (f, filename) = dump.write(mesh) # doctest: +GMSH
    >>> if parallelComm.Nproc == 1:
    ...     mesh = dump.read(filename, f) # doctest: +GMSH
    >>> print(mesh.globalNumberOfCells - len(numerix.nonzero(mesh.electrolyteMask)[0])) # doctest: +GMSH, +SERIAL
    150
    >>> print(400 < mesh.globalNumberOfCells < 800) # doctest: +GMSH
    True

    >>> from fipy.variables.cellVariable import CellVariable
    >>> var = CellVariable(mesh = mesh, value = 0.) # doctest: +GMSH

    >>> from fipy.terms.diffusionTerm import DiffusionTerm
    >>> eq = DiffusionTerm() # doctest: +GMSH

    >>> var.constrain(0., mesh.facesBottom) # doctest: +GMSH
    >>> var.constrain(domainHeight, mesh.facesTop) # doctest: +GMSH

    >>> eq.solve(var) # doctest: +GMSH

    Evaluate the result:

    >>> centers = mesh.cellCenters[1].copy() # doctest: +GMSH

    .. note:: the copy makes the array contiguous for inlining

    >>> localErrors = (centers - var)**2 / centers**2 # doctest: +GMSH
    >>> globalError = numerix.sqrt(numerix.sum(localErrors) / mesh.numberOfCells) # doctest: +GMSH
    >>> argmax = numerix.argmax(localErrors) # doctest: +GMSH
    >>> print(numerix.sqrt(localErrors[argmax]) < 0.051) # doctest: +GMSH
    1
    >>> print(globalError < 0.02) # doctest: +GMSH
    1

    """

    def __init__(self,
                 trenchDepth=None,
                 trenchSpacing=None,
                 boundaryLayerDepth=None,
                 cellSize=None,
                 aspectRatio=None,
                 angle=0.,
                 communicator=parallelComm):
        """

        `trenchDepth` - Depth of the trench.

        `trenchSpacing` - The distance between the trenches.

        `boundaryLayerDepth` - The depth of the hydrodynamic boundary
        layer.

        `cellSize` - The cell Size.

        `aspectRatio` - `trenchDepth` / `trenchWidth`

        `angle` - The angle for the taper of the trench.

        The trench mesh takes the parameters generally used to define
        a trench region and recasts then for the general
        `GapFillMesh`.

        """

        heightBelowTrench = cellSize * 10.

        heightAboveTrench = trenchDepth / 1.

        fineRegionHeight = heightBelowTrench + trenchDepth + heightAboveTrench
        transitionHeight = fineRegionHeight * 3.
        domainWidth = trenchSpacing / 2.
        domainHeight = heightBelowTrench + trenchDepth + boundaryLayerDepth

        super(TrenchMesh, self).__init__(cellSize=cellSize,
                                         desiredDomainWidth=domainWidth,
                                         desiredDomainHeight=domainHeight,
                                         desiredFineRegionHeight=fineRegionHeight,
                                         transitionRegionHeight=transitionHeight,
                                         communicator=parallelComm)

        trenchWidth = trenchDepth / aspectRatio

        x, y = self.cellCenters
        Y = (y - (heightBelowTrench + trenchDepth / 2))
        taper = numerix.tan(angle) * Y
        self.electrolyteMask = numerix.where(y > trenchDepth + heightBelowTrench,
                                             1,
                                             numerix.where(y < heightBelowTrench,
                                                           0,
                                                           numerix.where(x > trenchWidth / 2 + taper,
                                                                         0,
                                                                         1)))

    def __getstate__(self):
        dict = super(TrenchMesh, self).__getstate__()
        dict['electrolyteMask'] = self.electrolyteMask
        return dict

    def __setstate__(self, dict):
        self.electrolyteMask = dict['electrolyteMask']
        del dict['electrolyteMask']
        super(TrenchMesh, self).__setstate__(dict)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




__docformat__ = 'restructuredtext'

from .petscPreconditioner import PETScPreconditioner

__all__ = ["HYPREPreconditioner"]

class HYPREPreconditioner(PETScPreconditioner):
    """
    HYPRE preconditioner for :class:`~fipy.solvers.petsc.petscSolver.PETScSolver`.

    """

    pctype = "hypre"



r"""
Solves the Cahn-Hilliard problem in a 3D cube

>>> from fipy import CellVariable, Grid3D, Viewer, GaussianNoiseVariable, TransientTerm, DiffusionTerm, DefaultSolver
>>> from fipy.tools import numerix

The only difference from :mod:`examples.cahnHilliard.mesh2D` is the
declaration of ``mesh``.

>>> if __name__ == "__main__":
...     nx = ny = nz = 100
... else:
...     nx = ny = nz = 10
>>> mesh = Grid3D(nx=nx, ny=ny, nz=nz, dx=0.25, dy=0.25, dz=0.25)
>>> phi = CellVariable(name=r"$\phi$", mesh=mesh)

We start the problem with random fluctuations about :math:`\phi = 1/2`

>>> phi.setValue(GaussianNoiseVariable(mesh=mesh,
...                                    mean=0.5,
...                                    variance=0.01))

:term:`FiPy` doesn't plot or output anything unless you tell it to:

>>> if __name__ == "__main__":
...     viewer = Viewer(vars=(phi,), datamin=0., datamax=1.)

For :term:`FiPy`, we need to perform the partial derivative
:math:`\partial f/\partial \phi`
manually and then put the equation in the canonical
form by decomposing the spatial derivatives
so that each :class:`~fipy.terms.term.Term` is of a single, even order:

.. math::

   \frac{\partial \phi}{\partial t}
    = \nabla\cdot D a^2 \left[ 1 - 6 \phi \left(1 - \phi\right)\right] \nabla \phi- \nabla\cdot D \nabla \epsilon^2 \nabla^2 \phi.

:term:`FiPy` would automatically interpolate
``D * a**2 * (1 - 6 * phi * (1 - phi))``
onto the faces, where the diffusive flux is calculated, but we obtain
somewhat more accurate results by performing a linear interpolation from
``phi`` at cell centers to ``PHI`` at face centers.
Some problems benefit from non-linear interpolations, such as harmonic or
geometric means, and :term:`FiPy` makes it easy to obtain these, too.

>>> PHI = phi.arithmeticFaceValue
>>> D = a = epsilon = 1.
>>> eq = (TransientTerm()
...       == DiffusionTerm(coeff=D * a**2 * (1 - 6 * PHI * (1 - PHI)))
...       - DiffusionTerm(coeff=(D, epsilon**2)))

Because the evolution of a spinodal microstructure slows with time, we
use exponentially increasing time steps to keep the simulation
"interesting". The :term:`FiPy` user always has direct control over the
evolution of their problem.

>>> dexp = -5
>>> elapsed = 0.
>>> if __name__ == "__main__":
...     duration = 1000.
... else:
...     duration = 1e-2
>>> while elapsed < duration:
...     dt = min(100, numerix.exp(dexp))
...     elapsed += dt
...     dexp += 0.01
...     eq.solve(phi, dt=dt, solver=DefaultSolver(precon=None))
...     if __name__ == "__main__":
...         viewer.plot()

.. image:: /figures/examples/cahnHilliard/mesh3D.*
   :width: 90%
   :align: center
   :alt: snapshot of Cahn-Hilliard phase separation in 3D with cutaway
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')



__docformat__ = 'restructuredtext'

import platform

from fipy.terms.implicitSourceTerm import ImplicitSourceTerm
from fipy.variables.variable import Variable
from fipy.terms.explicitUpwindConvectionTerm import ExplicitUpwindConvectionTerm
from fipy.variables.surfactantConvectionVariable import SurfactantConvectionVariable
from fipy.terms.transientTerm import TransientTerm
from fipy.tests.doctestPlus import register_skipper
from fipy.solvers import solver

register_skipper(flag="NOTLINUXSCIPY",
                 test=lambda : platform.system() != "Linux" or solver != 'scipy',
                 why="`scipy` solvers on Linux fail intermittently: #575")

class AdsorbingSurfactantEquation(object):
    r"""

    The `AdsorbingSurfactantEquation` object solves the
    `SurfactantEquation` but with an adsorbing species from some bulk
    value. The equation that describes the surfactant adsorbing is
    given by,

    .. math::

       \dot{\theta} = J v \theta + k c (1 - \theta - \theta_{\text{other}}) - \theta c_{\text{other}} k_{\text{other}} - k^- \theta

    where :math:`\theta`, :math:`J`, :math:`v`, :math:`k`, :math:`c`,
    :math:`k^-` and :math:`n` represent the surfactant coverage, the curvature,
    the interface normal velocity, the adsorption rate, the concentration in the
    bulk at the interface, the consumption rate and an exponent of consumption,
    respectively. The :math:`\text{other}` subscript refers to another
    surfactant with greater surface affinity.

    The terms on the RHS of the above equation represent conservation of
    surfactant on a non-uniform surface, Langmuir adsorption, removal of
    surfactant due to adsorption of the other surfactant onto non-vacant sites
    and consumption of the surfactant respectively. The adsorption term is added
    to the source by setting :math:` S_c = k c (1 - \theta_{\text{other}})` and
    :math:`S_p = -k c`. The other terms are added to the source in a similar
    way.

    The following is a test case:

    >>> from fipy.variables.distanceVariable \
    ...     import DistanceVariable
    >>> from fipy import SurfactantVariable
    >>> from fipy.meshes import Grid2D
    >>> from fipy.tools import numerix
    >>> from fipy.variables.cellVariable import CellVariable
    >>> dx = .5
    >>> dy = 2.3
    >>> dt = 0.25
    >>> k = 0.56
    >>> initialValue = 0.1
    >>> c = 0.2

    >>> from fipy.meshes import Grid2D
    >>> from fipy import serialComm
    >>> mesh = Grid2D(dx = dx, dy = dy, nx = 5, ny = 1, communicator=serialComm)
    >>> distanceVar = DistanceVariable(mesh = mesh,
    ...                                value = (-dx*3/2, -dx/2, dx/2,
    ...                                          3*dx/2,  5*dx/2),
    ...                                hasOld = 1)
    >>> surfactantVar = SurfactantVariable(value = (0, 0, initialValue, 0, 0),
    ...                                    distanceVar = distanceVar)
    >>> bulkVar = CellVariable(mesh = mesh, value = (c, c, c, c, c))
    >>> eqn = AdsorbingSurfactantEquation(surfactantVar = surfactantVar,
    ...                                   distanceVar = distanceVar,
    ...                                   bulkVar = bulkVar,
    ...                                   rateConstant = k)
    >>> eqn.solve(surfactantVar, dt = dt)
    >>> answer = (initialValue + dt * k * c) / (1 + dt * k * c)
    >>> print(numerix.allclose(surfactantVar.interfaceVar,
    ...                  numerix.array((0, 0, answer, 0, 0))))
    1

    The following test case is for two surfactant variables. One has more
    surface affinity than the other.

    >>> from fipy.variables.distanceVariable \
    ...     import DistanceVariable
    >>> from fipy import SurfactantVariable
    >>> from fipy.meshes import Grid2D
    >>> dx = 0.5
    >>> dy = 2.73
    >>> dt = 0.001
    >>> k0 = 1.
    >>> k1 = 10.
    >>> theta0 = 0.
    >>> theta1 = 0.
    >>> c0 = 1.
    >>> c1 = 1.
    >>> totalSteps = 10
    >>> mesh = Grid2D(dx = dx, dy = dy, nx = 5, ny = 1, communicator=serialComm)
    >>> distanceVar = DistanceVariable(mesh = mesh,
    ...                                value = dx * (numerix.arange(5) - 1.5),
    ...                                hasOld = 1)
    >>> var0 = SurfactantVariable(value = (0, 0, theta0, 0, 0),
    ...                           distanceVar = distanceVar)
    >>> var1 = SurfactantVariable(value = (0, 0, theta1, 0, 0),
    ...                           distanceVar = distanceVar)
    >>> bulkVar0 = CellVariable(mesh = mesh, value = (c0, c0, c0, c0, c0))
    >>> bulkVar1 = CellVariable(mesh = mesh, value = (c1, c1, c1, c1, c1))

    >>> eqn0 = AdsorbingSurfactantEquation(surfactantVar = var0,
    ...                                    distanceVar = distanceVar,
    ...                                    bulkVar = bulkVar0,
    ...                                    rateConstant = k0)

    >>> eqn1 = AdsorbingSurfactantEquation(surfactantVar = var1,
    ...                                    distanceVar = distanceVar,
    ...                                    bulkVar = bulkVar1,
    ...                                    rateConstant = k1,
    ...                                    otherVar = var0,
    ...                                    otherBulkVar = bulkVar0,
    ...                                    otherRateConstant = k0)

    >>> for step in range(totalSteps):
    ...     eqn0.solve(var0, dt = dt)
    ...     eqn1.solve(var1, dt = dt)
    >>> answer0 = 1 - numerix.exp(-k0 * c0 * dt * totalSteps)
    >>> answer1 = (1 - numerix.exp(-k1 * c1 * dt * totalSteps)) * (1 - answer0)
    >>> print(numerix.allclose(var0.interfaceVar,
    ...                  numerix.array((0, 0, answer0, 0, 0)), rtol = 1e-2))
    1
    >>> print(numerix.allclose(var1.interfaceVar,
    ...                  numerix.array((0, 0, answer1, 0, 0)), rtol = 1e-2))
    1
    >>> dt = 0.1
    >>> for step in range(10):
    ...     eqn0.solve(var0, dt = dt)
    ...     eqn1.solve(var1, dt = dt)

    >>> x, y = mesh.cellCenters
    >>> check = var0.interfaceVar + var1.interfaceVar
    >>> answer = CellVariable(mesh=mesh, value=check)
    >>> answer[x==1.25] = 1.
    >>> print(check.allequal(answer))
    True

    The following test case is to fix a bug where setting the adsorption
    coefficient to zero leads to the solver not converging and an eventual
    failure.

    >>> var0 = SurfactantVariable(value = (0, 0, theta0, 0, 0),
    ...                           distanceVar = distanceVar)
    >>> bulkVar0 = CellVariable(mesh = mesh, value = (c0, c0, c0, c0, c0))

    >>> eqn0 = AdsorbingSurfactantEquation(surfactantVar = var0,
    ...                                    distanceVar = distanceVar,
    ...                                    bulkVar = bulkVar0,
    ...                                    rateConstant = 0)

    >>> eqn0.solve(var0, dt = dt)
    >>> eqn0.solve(var0, dt = dt)
    >>> answer = CellVariable(mesh=mesh, value=var0.interfaceVar)
    >>> answer[x==1.25] = 0.

    >>> print(var0.interfaceVar.allclose(answer))
    True

    The following test case is to fix a bug that allows the accelerator to
    become negative.

    >>> nx = 5
    >>> ny = 5
    >>> dx = 1.
    >>> dy = 1.
    >>> mesh = Grid2D(dx=dx, dy=dy, nx = nx, ny = ny, communicator=serialComm)
    >>> x, y = mesh.cellCenters

    >>> disVar = DistanceVariable(mesh=mesh, value=1., hasOld=True)
    >>> disVar[y < dy] = -1
    >>> disVar[x < dx] = -1
    >>> disVar.calcDistanceFunction() #doctest: +LSM

    >>> levVar = SurfactantVariable(value = 0.5, distanceVar = disVar)
    >>> accVar = SurfactantVariable(value = 0.5, distanceVar = disVar)

    >>> levEq = AdsorbingSurfactantEquation(levVar,
    ...                                     distanceVar = disVar,
    ...                                     bulkVar = 0,
    ...                                     rateConstant = 0)

    >>> accEq = AdsorbingSurfactantEquation(accVar,
    ...                                     distanceVar = disVar,
    ...                                     bulkVar = 0,
    ...                                     rateConstant = 0,
    ...                                     otherVar = levVar,
    ...                                     otherBulkVar = 0,
    ...                                     otherRateConstant = 0)

    >>> extVar = CellVariable(mesh = mesh, value = accVar.interfaceVar)

    >>> from fipy import TransientTerm, AdvectionTerm
    >>> advEq = TransientTerm() + AdvectionTerm(extVar)

    >>> dt = 0.1

    >>> for i in range(50):
    ...     disVar.calcDistanceFunction()
    ...     extVar.value = (numerix.array(accVar.interfaceVar))
    ...     disVar.extendVariable(extVar)
    ...     disVar.updateOld()
    ...     advEq.solve(disVar, dt = dt)
    ...     levEq.solve(levVar, dt = dt)
    ...     accEq.solve(accVar, dt = dt) #doctest: +LSM

    >>> # The following test fails sometimes on linux with scipy solvers
    >>> # See issue #575. We ignore for now.
    >>> print((accVar >= -1e-10).all()) #doctest: +NOTLINUXSCIPY
    True
    """
    def __init__(self,
                 surfactantVar = None,
                 distanceVar = None,
                 bulkVar = None,
                 rateConstant = None,
                 otherVar = None,
                 otherBulkVar = None,
                 otherRateConstant = None,
                 consumptionCoeff = None):
        """Create a `AdsorbingSurfactantEquation` object.

        Parameters
        ----------
        surfactantVar : ~fipy.variables.surfactantVariable.SurfactantVariable
            The `SurfactantVariable` to be solved for.
        distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
            The `DistanceVariable` that marks the interface.
        bulkVar : ~fipy.variables.cellVariable.CellVariable
            The value of the `surfactantVar` in the bulk.
        rateConstant : float
            The adsorption rate of the `surfactantVar`.
        otherVar : ~fipy.variables.surfactantVariable.SurfactantVariable
            Another `SurfactantVariable` with more surface affinity.
        otherBulkVar : ~fipy.variables.cellVariable.CellVariable
            The value of the `otherVar` in the bulk.
        otherRateConstant : float
            The adsorption rate of the `otherVar`.
        consumptionCoeff : float
            The rate that the `surfactantVar` is consumed during deposition.

        """

        self.eq = TransientTerm(coeff = 1) - ExplicitUpwindConvectionTerm(SurfactantConvectionVariable(distanceVar))

        self.dt = Variable(0.)
        mesh = distanceVar.mesh
        adsorptionCoeff = self.dt * bulkVar * rateConstant
        spCoeff = adsorptionCoeff * distanceVar._cellInterfaceFlag
        scCoeff = adsorptionCoeff * distanceVar.cellInterfaceAreas / mesh.cellVolumes

        self.eq += ImplicitSourceTerm(spCoeff) - scCoeff

        if otherVar is not None:
            otherSpCoeff = self.dt * otherBulkVar * otherRateConstant * distanceVar._cellInterfaceFlag
            otherScCoeff = -otherVar.interfaceVar * scCoeff

            self.eq += ImplicitSourceTerm(otherSpCoeff) - otherScCoeff

            vars = (surfactantVar, otherVar)
        else:
            vars = (surfactantVar,)

        total = 0
        for var in vars:
            total += var.interfaceVar
        maxVar = (total > 1) * distanceVar._cellInterfaceFlag

        val = distanceVar.cellInterfaceAreas / mesh.cellVolumes
        for var in vars[1:]:
            val -= distanceVar._cellInterfaceFlag * var

        spMaxCoeff = 1e20 * maxVar
        scMaxCoeff = spMaxCoeff * val * (val > 0)

        self.eq += ImplicitSourceTerm(spMaxCoeff) - scMaxCoeff - 1e-40

        if consumptionCoeff is not None:
            self.eq += ImplicitSourceTerm(consumptionCoeff)

    def solve(self, var, boundaryConditions=(), solver=None, dt=None):
        """Builds and solves the `AdsorbingSurfactantEquation`'s linear system once.

        Parameters
        ----------
        var : ~fipy.variables.surfactantVariable.SurfactantVariable
            A `SurfactantVariable` to be solved for.  Provides the initial
            condition, the old value and holds the solution on completion.
        solver : ~fipy.solvers.solver.Solver
            The iterative solver to be used to solve the linear system of
            equations.
        boundaryConditions : :obj:`tuple` of ~fipy.boundaryConditions.boundaryCondition.BoundaryCondition
        dt : float
            The time step size.

        """
        self.dt.setValue(dt)
        if solver is None:
            import fipy.solvers.solver
            from fipy.solvers import LinearPCGSolver
            solver = LinearPCGSolver()

        if type(boundaryConditions) not in (type(()), type([])):
            boundaryConditions = (boundaryConditions,)

        var.constrain(0, var.mesh.exteriorFaces)

        self.eq.solve(var,
                      boundaryConditions=boundaryConditions,
                      solver = solver,
                      dt=1.)

    def sweep(self, var, solver=None, boundaryConditions=(), dt=None, underRelaxation=None, residualFn=None):
        r"""Builds and solves the linear system once

        This method also recalculates and returns the
        residual as well as applying under-relaxation.

        Parameters
        ----------
        var : ~fipy.variables.cellVariable.CellVariable
            The `Variable` to be solved for.  Provides the initial
            condition, the old value and holds the solution on completion.
        solver : ~fipy.solvers.solver.Solver
            The iterative solver to be used to solve the linear system of
            equations.
        boundaryConditions : :obj`tuple` of ~fipy.boundaryConditions.boundaryCondition.BoundaryCondition
        dt : float
            The time step size.
        underRelaxation : float
            Usually a value between `0` and `1` or `None` in the case of no
            under-relaxation

        """
        self.dt.setValue(dt)
        if solver is None:
            from fipy.solvers import DefaultAsymmetricSolver
            solver = DefaultAsymmetricSolver()

        if type(boundaryConditions) not in (type(()), type([])):
            boundaryConditions = (boundaryConditions,)

        var.constrain(0, var.mesh.exteriorFaces)

        return self.eq.sweep(var, solver=solver, boundaryConditions=boundaryConditions, underRelaxation=underRelaxation, residualFn=residualFn, dt=1.)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



"""Test numeric implementation of the mesh
"""

__all__ = []

from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(docTestModuleNames=(
        'matplotlibViewer',
        'matplotlib1DViewer',
        'matplotlib2DViewer',
        'matplotlib2DGridViewer',
        'matplotlib2DContourViewer',
        'matplotlib2DGridContourViewer',
        'matplotlibStreamViewer',
        'matplotlibVectorViewer',
        ), base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



import re

import numpy

scanf_e = r"[-+]?(\d+(\.\d*)?|\d*\.\d+)([eE][-+]?\d+)?"

reCPU = re.compile("cpu time: (%s) s / step / cell" % scanf_e)
reRSZ = re.compile("max resident memory: (%s) B / cell" % scanf_e)
reVSZ = re.compile("max virtual memory: (%s) B / cell" % scanf_e)

def monitor(p):
    r = "".join(p.communicate()[0])

    cpu = reCPU.search(r, re.MULTILINE)
    rsz = reRSZ.search(r, re.MULTILINE)
    vsz = reVSZ.search(r, re.MULTILINE)

    def numOrNaN(m, g=1):
        if m is None:
            return numpy.NaN
        else:
            return float(m.group(g))

    return (numOrNaN(cpu),
            numOrNaN(rsz),
            numOrNaN(vsz))



"""

The `gapFillMesh` function glues 3 meshes together to form a composite
mesh. The first mesh is a `Grid2D` object that is fine and deals with
the area around the trench or via. The second mesh is a `Gmsh2D`
object that forms a transition mesh from a fine to a course
region. The third mesh is another `Grid2D` object that forms the
boundary layer. This region consists of very large elements and is
only used for the diffusion in the boundary layer.

"""

__docformat__ = 'restructuredtext'

from packaging.version import Version

from fipy.meshes import Gmsh2D
from fipy.meshes.gmshMesh import _gmshVersion
from fipy.meshes import Grid2D
from fipy.tools import serialComm
from fipy.tools import parallelComm

class GapFillMesh(Gmsh2D):
    """
    The following test case tests for diffusion across the domain.
    >>> domainHeight = 5.
    >>> mesh = GapFillMesh(transitionRegionHeight = 2.,
    ...                    cellSize = 0.1,
    ...                    desiredFineRegionHeight = 1.,
    ...                    desiredDomainHeight = domainHeight,
    ...                    desiredDomainWidth = 1.) # doctest: +GMSH

    >>> import fipy.tools.dump as dump
    >>> (f, filename) = dump.write(mesh) # doctest: +GMSH
    >>> if parallelComm.Nproc == 1:
    ...     mesh = dump.read(filename, f) # doctest: +GMSH

    >>> print(136 < mesh.globalNumberOfCells < 300) # doctest: +GMSH
    True

    >>> from fipy.variables.cellVariable import CellVariable
    >>> var = CellVariable(mesh = mesh) # doctest: +GMSH

    >>> from fipy.terms.diffusionTerm import DiffusionTerm
    >>> eq = DiffusionTerm()

    >>> var.constrain(0., mesh.facesBottom) # doctest: +GMSH
    >>> var.constrain(domainHeight, mesh.facesTop) # doctest: +GMSH

    >>> eq.solve(var) # doctest: +GMSH

    Evaluate the result:

    >>> centers = mesh.cellCenters[1].copy() # doctest: +GMSH

    .. note:: the copy makes the array contiguous for inlining

    >>> localErrors = (centers - var)**2 / centers**2 # doctest: +GMSH
    >>> from fipy.tools import numerix
    >>> globalError = numerix.sqrt(numerix.sum(localErrors) / mesh.numberOfCells) # doctest: +GMSH
    >>> argmax = numerix.argmax(localErrors) # doctest: +GMSH

    >>> print(numerix.sqrt(localErrors[argmax]) < 0.1) # doctest: +GMSH
    1
    >>> print(globalError < 0.05) # doctest: +GMSH
    1

    """
    def __init__(self,
                 cellSize=None,
                 desiredDomainWidth=None,
                 desiredDomainHeight=None,
                 desiredFineRegionHeight=None,
                 transitionRegionHeight=None,
                 communicator=parallelComm):

        """
        Arguments:

        `cellSize` - The cell size in the fine grid around the trench.

        `desiredDomainWidth` - The desired domain width.

        `desiredDomainHeight` - The total desired height of the
        domain.

        `desiredFineRegionHeight` - The desired height of the in the
        fine region around the trench.

        `transitionRegionHeight` - The height of the transition region.
        """

        # Calculate the fine region cell counts.
        nx = int(desiredDomainWidth / cellSize)
        ny = int(desiredFineRegionHeight / cellSize)

        # Calculate the actual mesh dimensions
        actualFineRegionHeight = ny * cellSize
        actualDomainWidth = nx * cellSize
        boundaryLayerHeight = desiredDomainHeight - actualFineRegionHeight - transitionRegionHeight
        numberOfBoundaryLayerCells = int(boundaryLayerHeight / actualDomainWidth)

        # Build the fine region mesh.
        self.fineMesh = Grid2D(nx=nx, ny=ny, dx=cellSize, dy=cellSize, communicator=serialComm)

        if _gmshVersion() < Version("2.7"):
            # kludge: must offset cellSize by `eps` to work properly
            eps = float(cellSize)/(nx * 10)
        else:
            eps = 0.

        super(GapFillMesh, self).__init__("""
        ny       = %(ny)g;
        cellSize = %(cellSize)g - %(eps)g;
        height   = %(actualFineRegionHeight)g;
        width    = %(actualDomainWidth)g;
        boundaryLayerHeight = %(boundaryLayerHeight)g;
        transitionRegionHeight = %(transitionRegionHeight)g;
        numberOfBoundaryLayerCells = %(numberOfBoundaryLayerCells)g;

        Point(1) = {0, 0, 0, cellSize};
        Point(2) = {width, 0, 0, cellSize};
        Line(3) = {1, 2};

        Point(10) = {0, height, 0, cellSize};
        Point(11) = {width, height, 0, cellSize};
        Point(12) = {0, height + transitionRegionHeight, 0, width};
        Point(13) = {width, height + transitionRegionHeight, 0, width};
        Line(14) = {10,11};
        Line(15) = {11,13};
        Line(16) = {13,12};
        Line(17) = {12,10};
        Line Loop(18) = {14, 15, 16, 17};
        Plane Surface(19) = {18};

        Extrude{0, height, 0} {
            Line{3}; Layers{ ny }; Recombine;}

        Line(100) = {12, 13};
        Extrude{0, boundaryLayerHeight, 0} {
            Line{100}; Layers{ numberOfBoundaryLayerCells }; Recombine;}
        """ % locals(), communicator=communicator)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



__docformat__ = 'restructuredtext'

from fipy.terms.abstractDiffusionTerm import _AbstractDiffusionTerm
from fipy.tools import numerix

__all__ = ["DiffusionTermCorrection"]

class DiffusionTermCorrection(_AbstractDiffusionTerm):

    def _getNormals(self, mesh):
        return mesh._faceCellToCellNormals

    def _treatMeshAsOrthogonal(self, mesh):
        return mesh._isOrthogonal



__docformat__ = 'restructuredtext'

from fipy.solvers.petsc.petscKrylovSolver import PETScKrylovSolver

__all__ = ["LinearBicgSolver"]

class LinearBicgSolver(PETScKrylovSolver):

    """Interface to the biconjugate gradient solver (:term:`BiCG`) in
    :ref:`PETSC`.
    """
      
    solver = 'bicg'
